/*
 * UCB-BF707 N25Q128 DPIAhCo
 *
 * Copyright(C) 2016 Kaneko System Co., Ltd
 *
 * Version : 0.0.1
 *
 * [History]
 *          Date       Author                     Version     Description
 * ---------------------------------------------------------------------------
 * CREATE : 01/Aug/2016 Shinya Kaneko             v.0.0.1     First release
 *
 * [Note]
 *
 */
#include <sys/platform.h>
#include <string.h>
#include "../common/flash.h"

/* STATUS bits */

#define STATUS1_BUSY        0x01
#define STATUS1_WEL         0x02

/* Various sizes */
#define PAGE_SIZE           0x100
#define SECTOR_SIZE         0x10000
#define NUM_SECTORS         1024

static int n25q128a_disable_quad_mode(const struct flash_info *fi);
static int n25q128a_read_status(const struct flash_info *fi, uint8_t *status, int n);

static void bf70x_spi2_init(void)
{
    *pREG_PORTB_FER_SET = BITM_PORT_FER_SET_PX15 | BITM_PORT_FER_SET_PX12 | BITM_PORT_FER_SET_PX11 | BITM_PORT_FER_SET_PX10;
    *pREG_PORTB_MUX &= ~(BITM_PORT_MUX_MUX15 | BITM_PORT_MUX_MUX12 | BITM_PORT_MUX_MUX11 | BITM_PORT_MUX_MUX10);

    *pREG_SPI2_IMSK = 0U;
    *pREG_SPI2_CTL = 0U;
    *pREG_SPI2_TXCTL = 0U;
    *pREG_SPI2_RXCTL = 0U;
    *pREG_DMA8_CFG = 0U;
    *pREG_DMA9_CFG = 0U;
    ssync();

    *pREG_SPI2_STAT =
        BITM_SPI_STAT_MMAE | BITM_SPI_STAT_MMRE | BITM_SPI_STAT_MMWE |
        BITM_SPI_STAT_TF | BITM_SPI_STAT_RF | BITM_SPI_STAT_TS | BITM_SPI_STAT_RS |
        BITM_SPI_STAT_MF | BITM_SPI_STAT_TC | BITM_SPI_STAT_TUR | BITM_SPI_STAT_ROR;
    *pREG_SPI2_CLK = 2; /* 25MHz */
    *pREG_SPI2_DLY = 0U;
    *pREG_SPI2_SLVSEL =
        ENUM_SPI_SLVSEL_SSEL1_LO | ENUM_SPI_SLVSEL_SSEL1_EN |
        ENUM_SPI_SLVSEL_SSEL2_LO | ENUM_SPI_SLVSEL_SSEL2_EN |
        ENUM_SPI_SLVSEL_SSEL3_LO | ENUM_SPI_SLVSEL_SSEL3_EN |
        ENUM_SPI_SLVSEL_SSEL4_LO | ENUM_SPI_SLVSEL_SSEL4_EN |
        ENUM_SPI_SLVSEL_SSEL5_LO | ENUM_SPI_SLVSEL_SSEL5_EN |
        ENUM_SPI_SLVSEL_SSEL6_LO | ENUM_SPI_SLVSEL_SSEL6_EN |
        ENUM_SPI_SLVSEL_SSEL7_LO | ENUM_SPI_SLVSEL_SSEL7_EN;

    *pREG_DMA8_STAT = BITM_DMA_STAT_PIRQ | BITM_DMA_STAT_IRQDONE | BITM_DMA_STAT_IRQERR;
    *pREG_DMA8_XCNT = 0;
    *pREG_DMA8_XMOD = sizeof(uint8_t);
    *pREG_DMA8_CFG =
        ENUM_DMA_CFG_PDAT_NOTFWD | ENUM_DMA_CFG_ADDR1D | ENUM_DMA_CFG_NO_COPY | ENUM_DMA_CFG_TOV_DIS |
        ENUM_DMA_CFG_NO_TRIG | ENUM_DMA_CFG_XCNT_INT | /* ENUM_DMA_CFG_FETCH07 | */ ENUM_DMA_CFG_NO_TRGWAIT |
        ENUM_DMA_CFG_STOP | ENUM_DMA_CFG_MSIZE01 | ENUM_DMA_CFG_PSIZE01 | ENUM_DMA_CFG_LD_STARTADDR |
        ENUM_DMA_CFG_NO_SYNC | ENUM_DMA_CFG_READ;

    *pREG_DMA9_STAT = BITM_DMA_STAT_PIRQ | BITM_DMA_STAT_IRQDONE | BITM_DMA_STAT_IRQERR;
    *pREG_DMA9_XCNT = 0;
    *pREG_DMA9_XMOD = sizeof(uint8_t);
    *pREG_DMA9_CFG =
        ENUM_DMA_CFG_PDAT_NOTFWD | ENUM_DMA_CFG_ADDR1D | ENUM_DMA_CFG_NO_COPY | ENUM_DMA_CFG_TOV_DIS |
        ENUM_DMA_CFG_NO_TRIG | ENUM_DMA_CFG_XCNT_INT | /* ENUM_DMA_CFG_FETCH07 | */ ENUM_DMA_CFG_NO_TRGWAIT |
        ENUM_DMA_CFG_STOP | ENUM_DMA_CFG_MSIZE01 | ENUM_DMA_CFG_PSIZE01 | ENUM_DMA_CFG_LD_STARTADDR |
        ENUM_DMA_CFG_NO_SYNC | ENUM_DMA_CFG_WRITE;

    *pREG_SPI2_CTL =
        ENUM_SPI_CTL_MIO_DIS |
        ENUM_SPI_CTL_FAST_EN |
        ENUM_SPI_CTL_MSB_FIRST |
        ENUM_SPI_CTL_SIZE08 |
        ENUM_SPI_CTL_MISO_EN |
        ENUM_SPI_CTL_SW_SSEL |
        ENUM_SPI_CTL_MASTER;
}

static void bf70x_spi2_close(void)
{
    *pREG_PORTB_FER_CLR =
        BITM_PORT_FER_SET_PX15 | BITM_PORT_FER_SET_PX12 |
        BITM_PORT_FER_SET_PX11 | BITM_PORT_FER_SET_PX10;

    *pREG_DMA8_CFG = 0U;
    *pREG_DMA9_CFG = 0U;
    *pREG_SPI2_IMSK = 0U;
    *pREG_SPI2_CTL = 0U;
    *pREG_SPI2_TXCTL = 0U;
    *pREG_SPI2_RXCTL = 0U;
    *pREG_SPI2_STAT =
        BITM_SPI_STAT_MMAE | BITM_SPI_STAT_MMRE | BITM_SPI_STAT_MMWE |
        BITM_SPI_STAT_TF | BITM_SPI_STAT_RF | BITM_SPI_STAT_TS | BITM_SPI_STAT_RS |
        BITM_SPI_STAT_MF | BITM_SPI_STAT_TC | BITM_SPI_STAT_TUR | BITM_SPI_STAT_ROR;
}

static void bf70x_spi2_write(const uint8_t* p_tx_data, int32_t count)
{
    int32_t txc;

    *pREG_SPI2_STAT =
        BITM_SPI_STAT_TF | BITM_SPI_STAT_RF | BITM_SPI_STAT_TS | BITM_SPI_STAT_RS |
        BITM_SPI_STAT_MF | BITM_SPI_STAT_TC | BITM_SPI_STAT_TUR | BITM_SPI_STAT_ROR;

    *pREG_SPI2_TWC = count;
    *pREG_SPI2_TXCTL = ENUM_SPI_TXCTL_UWM_DIS | ENUM_SPI_TXCTL_RWM_75 | ENUM_SPI_TXCTL_TDR_NF | ENUM_SPI_TXCTL_TWC_EN | ENUM_SPI_TXCTL_TTI_EN | ENUM_SPI_TXCTL_TX_EN;
    *pREG_DMA8_ADDRSTART = (void*)p_tx_data;
    *pREG_DMA8_XCNT = count;
    *pREG_DMA8_CFG |= ENUM_DMA_CFG_EN;
    while((*pREG_DMA8_STAT & BITM_DMA_STAT_IRQDONE) == ENUM_DMA_STAT_NO_IRQ) { ssync(); }
    while((*pREG_SPI2_STAT & BITM_SPI_STAT_TF) == ENUM_SPI_STAT_TF_LO) { ssync(); }
    while((*pREG_SPI2_STAT & BITM_SPI_STAT_SPIF) == ENUM_SPI_STAT_SPIF_LO) { ssync(); }
    *pREG_SPI2_TXCTL = 0U;
    *pREG_DMA8_CFG &= ~BITM_DMA_CFG_EN;
}

static void bf70x_spi2_read(uint8_t* p_rx_data, int32_t count)
{
    int32_t rxc;

    *pREG_SPI2_STAT =
        BITM_SPI_STAT_TF | BITM_SPI_STAT_RF | BITM_SPI_STAT_TS | BITM_SPI_STAT_RS |
        BITM_SPI_STAT_MF | BITM_SPI_STAT_TC | BITM_SPI_STAT_TUR | BITM_SPI_STAT_ROR;

    *pREG_DMA9_ADDRSTART = p_rx_data;
    *pREG_DMA9_XCNT = count;
    *pREG_DMA9_CFG |= ENUM_DMA_CFG_EN;
    *pREG_SPI2_RXCTL =
        ENUM_SPI_RXCTL_UWM_75 | ENUM_SPI_RXCTL_RWM_50 | ENUM_SPI_RXCTL_RDR_NE |
        ENUM_SPI_RXCTL_RTI_EN | ENUM_SPI_RXCTL_RX_EN;
    while((*pREG_DMA9_STAT & BITM_DMA_STAT_IRQDONE) == ENUM_DMA_STAT_NO_IRQ) { ssync(); }
    *pREG_SPI2_RXCTL = 0;
    *pREG_DMA9_CFG &= ~BITM_DMA_CFG_EN;
}

static void bf70x_spi2_cs_assert(uint32_t bits)
{
    *pREG_SPI2_CTL |= BITM_SPI_CTL_EN;
    *pREG_SPI2_SLVSEL &= ~bits;
    ssync();
}

static void bf70x_spi2_cs_deassert(uint32_t bits)
{
    *pREG_SPI2_SLVSEL |= bits;
    *pREG_SPI2_CTL &= ~BITM_SPI_CTL_EN;
    ssync();
}


static int n25q128a_open(struct flash_info *fi)
{
    uint8_t l_config;

    bf70x_spi2_init();

    fi->size = 16 * 1024 * 1024;
    fi->number_of_regions = 1;
    fi->erase_block_regions = malloc(fi->number_of_regions * sizeof (struct erase_block_region));

    if (fi->erase_block_regions == NULL)
        return -1;

    fi->erase_block_regions[0].block_size = SECTOR_SIZE;
    fi->erase_block_regions[0].number_of_blocks = NUM_SECTORS;
    return 0;
}

static int n25q128a_close(struct flash_info *fi)
{
    bf70x_spi2_close();
    return 0;
}

static int n25q128a_write_enable(const struct flash_info *fi)
{
    uint8_t l_cmd[1];
    uint8_t l_status;

    do
    {
        l_cmd[0] = 0x06U;
        bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
        bf70x_spi2_write(&l_cmd[0], 1);
        bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

        n25q128a_read_status(fi, &l_status, 1);
    } while((l_status & 0x02U) == 0U);

    return 0;
}

static int n25q128a_write_disable(const struct flash_info *fi)
{
    uint8_t l_cmd[2];

    l_cmd[0] = 0x04U;
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    bf70x_spi2_write(&l_cmd[0], 1);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

    return 0;
}

static int n25q128a_read_mid_did(struct flash_info *fi, uint8_t *mid, uint8_t *did)
{
    uint8_t l_cmd[4];

    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    l_cmd[0] = 0x9EU;
    bf70x_spi2_write(&l_cmd[0], 1);
    bf70x_spi2_read(&l_cmd[0], 4);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

    *mid = l_cmd[0];
    *did = l_cmd[1];
    return 0;
}

static int n25q128a_read_uid(const struct flash_info *fi, uint64_t *uid)
{
    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    *uid = 0U;
    return 0;
}

static int n25q128a_read_jedec_id(const struct flash_info *fi, uint8_t *mid,
        uint8_t *memory_type_id, uint8_t *capacity_id)
{
    uint8_t l_cmd[4];

    /* Only for STANDARD mode */
    if (fi->current_mode != STANDARD)
        return -1;

    /* :Single Modep */
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    l_cmd[0] = 0x9E;
    bf70x_spi2_write(&l_cmd[0], 1);
    bf70x_spi2_read(&l_cmd[0], 3);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

    *mid = l_cmd[0];
    *memory_type_id = l_cmd[1];
    *capacity_id = l_cmd[2];
    return 0;
}

static int n25q128a_read_status(const struct flash_info *fi, uint8_t *status, int n)
{
    uint8_t l_cmd[2];

    l_cmd[0] = 0x05;
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    bf70x_spi2_write(&l_cmd[0], 1);
    bf70x_spi2_read(&l_cmd[1], 1);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);
    *status = l_cmd[1];
    return 0;
}

static int n25q128a_read_status1(const struct flash_info *fi, uint8_t *status)
{
    return n25q128a_read_status(fi, status, 1);
}

static int n25q128a_read_status2(const struct flash_info *fi, uint8_t *status)
{
    return n25q128a_read_status(fi, status, 2);
}

static int n25q128a_read_flag(const struct flash_info *fi, uint8_t *flag)
{
    uint8_t l_cmd[2];

    l_cmd[0] = 0x70;
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    bf70x_spi2_write(&l_cmd[0], 1);
    bf70x_spi2_read(&l_cmd[1], 1);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

    *flag = l_cmd[1];
    return 0;
}


static int n25q128a_not_busy(const uint8_t status)
{
    return (status & STATUS1_BUSY) ? 0 : 1;
}

static int n25q128a_wait_ready(const struct flash_info *fi)
{
    uint8_t status;

    do
    {
        n25q128a_read_status1(fi, &status);
    } while(!n25q128a_not_busy(status));

    return 0;
}

static int n25q128a_write_status(const struct flash_info *fi, uint8_t status1, uint8_t status2)
{
    return 0;
}

static int n25q128a_enable_quad_mode(const struct flash_info *fi)
{
    return 0;
}

static int n25q128a_disable_quad_mode(const struct flash_info *fi)
{
    return 0;
}

static int n25q128a_read(const struct flash_info *fi, uint32_t addr, uint8_t *buf, int size)
{
    uint8_t l_cmd[4];

    l_cmd[0] = 0x03;
    l_cmd[1] = (addr >> 16) & 0xFFU;
    l_cmd[2] = (addr >> 8) & 0xFFU;
    l_cmd[3] = (addr >> 0) & 0xFFU;
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    bf70x_spi2_write(l_cmd, 1+3);
    bf70x_spi2_read(buf, size);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);
    return 0;
}

static int n25q128a_erase_1(const struct flash_info *fi,
        uint32_t addr, erase_type_t erase_type)
{
    uint8_t l_cmd[4];
    uint8_t l_flag;

    n25q128a_write_enable(fi);

    if(erase_type == ERASE_CHIP)
    {
        l_cmd[0] = 0xC7;
        bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
        bf70x_spi2_write(&l_cmd[0], 1);
        bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);
    }
    else
    {
        l_cmd[0] = 0xD8;
        l_cmd[1] = (addr >> 16) & 0xFFU;
        l_cmd[2] = (addr >> 8) & 0xFFU;
        l_cmd[3] = (addr >> 0) & 0xFFU;
        bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
        bf70x_spi2_write(&l_cmd[0], 1+3);
        bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);
    }

    n25q128a_wait_ready(fi);
    do
    {
        n25q128a_read_flag(fi, &l_flag);
    } while((l_flag & 0x01U) != 0x00U);

    return 0;
}

static int n25q128a_erase(const struct flash_info *fi, uint32_t addr, int size)
{
    size += addr % SECTOR_SIZE;
    addr -= addr % SECTOR_SIZE;

    while(size > 0)
    {
        n25q128a_erase_1(fi, addr, ERASE_64KB);
        addr += SECTOR_SIZE;
        size -= SECTOR_SIZE;
    }

    return 0;
}

static int n25q128a_erase_chip(const struct flash_info *fi)
{
    return n25q128a_erase_1(fi, 0, ERASE_CHIP);
}

static int n25q128a_page_program(struct flash_info *fi,
        uint32_t addr, const uint8_t *buf, int size)
{
    uint8_t l_cmd[4];
    uint8_t l_flag;

    n25q128a_write_enable(fi);

    l_cmd[0] = 0x02;
    l_cmd[1] = (addr >> 16) & 0xFFU;
    l_cmd[2] = (addr >> 8) & 0xFFU;
    l_cmd[3] = (addr >> 0) & 0xFFU;
    bf70x_spi2_cs_assert(BITM_SPI_SLVSEL_SSEL1);
    bf70x_spi2_write(&l_cmd[0], 4);
    bf70x_spi2_write(buf, size);
    bf70x_spi2_cs_deassert(BITM_SPI_SLVSEL_SSEL1);

    n25q128a_wait_ready(fi);
    do
    {
        n25q128a_read_flag(fi, &l_flag);
    } while((l_flag & 0x01U) != 0x00U);
    return 0;
}

/* Program arbitrary size */
static int n25q128a_program(struct flash_info *fi,
        uint32_t addr, const uint8_t *buf, int size)
{
    int program_size;

    while (size > 0)
    {
        if (addr % PAGE_SIZE != 0)
            program_size = ((PAGE_SIZE - addr % PAGE_SIZE) < size
                    ? (PAGE_SIZE - addr % PAGE_SIZE) : size);
        else if (size >= PAGE_SIZE)
            program_size = PAGE_SIZE;
        else
            program_size = size;

        n25q128a_page_program(fi, addr, buf, program_size);

        addr += program_size;
        buf += program_size;
        size -= program_size;
    }

    return 0;
}

static int n25q128a_reset(const struct flash_info *fi)
{
    /* Not supported */
    return -1;
}

struct flash_info n25q128a_info =
{
    "n25q128a",  /* name */
    "Micron",   /* mname */
    0x20,       /* manufacturer ID */
    0xBB,       /* device ID */
    0x20,       /* memory type ID */
    0x18,       /* capacity ID */
    STANDARD, /* supported modes */


    /* The following three fields are better be
       initialized in flash_open functions.  */
    0,          /* size in Byte */
    0,          /* number_of_regions */
    NULL,       /* erase_block_regions */
    108,        /* max_freq */
    108,        /* max_quad_freq */
    50,         /* max_read_data_freq */

    0,          /* cpha */
    0,          /* cpol */
    0,          /* lsb_first */
    0,          /* start_on_mosi */

    0,          /* frequency */
    STANDARD,   /* current_mode */

    n25q128a_open,
    n25q128a_close,

    n25q128a_read_mid_did,
    n25q128a_read_uid,
    n25q128a_read_jedec_id,
    n25q128a_read_status1,
    n25q128a_read_status2,
    n25q128a_write_status,
    n25q128a_enable_quad_mode,
    n25q128a_disable_quad_mode,
    n25q128a_read,
    n25q128a_write_enable,
    n25q128a_write_disable,
    n25q128a_erase,
    n25q128a_erase_chip,
    n25q128a_program,
    n25q128a_reset,
};
